#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#ifndef NO_READLINE
# include <readline/readline.h>
# include <readline/history.h>
#endif /* !NO_READLINE */
#include <stutter.h>
#include <assert.h>

/* No readline support on WinCE.
 * FIXME: no console support, either. --dig */
#ifndef NO_READLINE

varctx *completion_ctx;

char * word_generator(const char *text, int state)
{
	static int slot, len;
	static s_var *elem;
	char *name;

	if(!state) {
		slot = 0;
		elem = NULL;
		len = strlen(text);
	}

	while(slot < completion_ctx->ns) {
		if(!elem) elem = completion_ctx->hash[slot];
		if(!elem) {
			slot++;
			continue;
		}

		name = get_symbolname(elem->sdbm, elem->id);
		elem = elem->next;
		if(!elem) slot++;
		if(strncasecmp(name, text, len) == 0)
			return strdup(name);
	}

	return NULL;
}

char ** stutter_completion(const char *text, int start, int end)
{
	return rl_completion_matches(text, word_generator);
}

void shell()
{
	s_obj *obj;
	s_mark mark;
	varctx *global;
	char *buf, *c;
	read_state *rs;

	char *hist;
	int histbuf, histlen;
	int beg = 1;

	/* set up readline */
	rl_readline_name = "stutter";
	rl_attempted_completion_function = stutter_completion;




	rs = create_read_state();

	histbuf = 256;
	hist = (char *) malloc(histbuf);
	assert(hist != NULL);
	histlen = 0;
	hist[0] = 0;

	printf("Welcome to STUTTER %s\n\n", stutter_version);

	completion_ctx = global = varctx_create(NULL, 1024);
	register_builtins(global);
#ifdef NEWPARSER
	obj = register_syntax(global);
	if(obj->type == ST_ERROR) {
		printobj(global, obj);
		printf("\n");
	}
#endif

	varctx_set(global, "*ARGV*", create_nil());
	varctx_set(global, "$0", create_string("shell"));

	while((buf = readline(rs->stack->type != ST_NIL ? "" : "> "))) {
		obj = NULL;
		for(c = buf; ; c++) {
			if(*c && beg && !isspace(*c))
				beg = 0;

			if(!beg) {
				if(histlen == histbuf - 1) {
					hist = (char *) realloc(hist, histbuf *= 2);
					assert(hist != NULL);
				}
				hist[histlen] = *c ? *c : '\n';
				hist[++histlen] = 0;
			}

			if(*c) {
				obj = read_ch(rs, *c);
			} else {
				obj = read_ch(rs, '\n');
			}
			if(obj) {
				if(obj->type == ST_ERROR && obj->d.error.code == SE_PARSE) {
					for(c++; *c; c++) {
						if(histlen == histbuf - 1) {
							hist = (char *) realloc(hist, histbuf *= 2);
							assert(hist != NULL);
						}
						hist[histlen] = *c ? *c : '\n';
						hist[++histlen] = 0;
					}

					add_history(hist);
					histlen = 0;
					printobj(global, obj);
					stream_stdout_write(global, '\n');
					read_reset(rs);
					break;
				}

				add_history(hist);
				histlen = 0;
				if(obj->type == ST_ERROR) break;
				mark = gc_prot_mark();
				obj = eval(global, compile(global, obj));
				printobj(global, obj);
				stream_stdout_write(global, '\n');
				gc_prot_free(mark, NULL);
				obj = NULL;
				beg = 1;
			}
			if(!*c) break;
		}
	}

	free(hist);
	free_read_state(rs);
	varctx_free(global);
}

#else  /* !NO_READLINE */

/* this shell() reads from stdin */
void shell()
{
	s_obj *obj;
	s_mark mark;
	varctx *global;
	int c;
	read_state *rs;

	rs = create_read_state();

	printf("Welcome to STUTTER %s\n\n> ", stutter_version);

	global = varctx_create(NULL, 1024);
	register_builtins(global);
#ifdef NEWPARSER
	obj = register_syntax(global);
	if(obj->type == ST_ERROR) {
		printobj(global, obj);
		printf("\n");
	}
#endif

	varctx_set(global, "*ARGV*", create_nil());
	varctx_set(global, "*NO-READLINE*", create_t());
	varctx_set(global, "$0", create_string("shell"));

	while((c = getchar())) {
		obj = read_ch(rs, c);

		if(obj) {
			if(obj->type == ST_ERROR) {
				printobj(global, obj);
				putchar('\n');
				read_reset(rs);
			} else {
				mark = gc_prot_mark();
				obj = eval(global, compile(global, obj));
				printobj(global, obj);
				putchar('\n');
				gc_prot_free(mark, NULL);
			}
		}

		if(c == '\n') fputs("> ", stdout);
		if(c == EOF) break;
	}

	free_read_state(rs);
	varctx_free(global);
}


#endif /* !NO_READLINE */

int main(int argc, char *argv[])
{
	int res, i, sh = 1;
	s_obj *obj;
	s_obj *sargv;
	varctx *global;

	while((res = getopt(argc, argv, "hvG:ge")) != -1) {
		if(res == '?' || res == 'h') {
			printf("STUTTER %s\n\n", stutter_version);
			printf(" -v       print interpreter version\n");
			printf(" -h       print this help screen\n");
			printf(" -G <n>   set auto-GC interval to per N objects allocated\n");
			printf(" -g       verbose GC\n");
			printf(" -e       disable readline interface (parse from stdin)\n");
			return 0;
		}

		if(res == 'e')
			sh = 0;
		if(res == 'g')
			GC_VERBOSE = 1;
		if(res == 'G') {
			GC_ALLOC_LIMIT = atoi(optarg);
			if(GC_ALLOC_LIMIT < 0) {
				GC_ALLOC_LIMIT = 0;
				printf("GC_ALLOC_LIMIT cannot be smaller than 0\n");
			}
		}

		if(res == 'v') {
			printf("%s\n", stutter_version);
			return 0;
		}
	}

	stutter_init();
	atexit(stutter_shutdown);

	if(optind == argc && sh) {
		shell();
		return 0;
	}

	global = varctx_create(NULL, 1024);
	register_builtins(global);
#ifdef NEWPARSER
	obj = register_syntax(global);
	if(obj->type == ST_ERROR) {
		printobj(global, obj);
		printf("\n");
	}
#endif


	if(optind == argc) {
		varctx_set(global, "*ARGV*", create_nil());
		varctx_set(global, "$0", create_string("stdin"));
		evalfp(global, stdin);
	} else {
		sargv = create_nil();
		obj = NULL;
		for(i=optind+1; i<argc; i++) {
			if(!obj) {
				obj = sargv = create_cons(create_string(argv[i]), create_nil());
			} else {
				obj->d.cons.cdr = create_cons(create_string(argv[i]), create_nil());
				obj = obj->d.cons.cdr;
			}
		}
		varctx_set(global, "*ARGV*", sargv);

		if(optind < argc) {
			varctx_set(global, "$0", create_string(argv[optind]));
			obj = evalfile(global, argv[optind]);
			if(obj->type == ST_ERROR) {
				printobj(global, obj);
				stream_stdout_write(global, '\n');
				varctx_free(global);
				gc_prot_free(0, NULL);
				gc_sweep(NULL);
				return 1;
			}
		}
	}

	varctx_free(global);

	return 0;
}

