#ifndef _ST_EVAL_H
#define _ST_EVAL_H

#ifdef __POCKETPC__
#define strdup _strdup
#endif

#include <stdio.h>
#include "types.h"
#include "memory.h"

#ifndef FUNC_LOCALCTX_SIZE
#define FUNC_LOCALCTX_SIZE 32 /* number of slots in the local variable contexts of functions */
#endif
#ifndef FUNC_MAXPARMS
#define FUNC_MAXPARMS 32 /* maximum number of parameters a function can take. just use a list.. */
#endif

extern int _stutter_trace; /* >0 if tracing is enabled */
extern int _stutter_depth; /* trace depth */

extern char *_stutter_file; /* current file being parsed */

extern char *_stutter_lastcall_file; /* location of last function call */
extern int _stutter_lastcall_line;

extern int _stutter_parsed_line; /* line number currently being parsed */

extern int recursion_limit; /* levels of recursion left */

extern FILE *yyin;

extern s_obj *_stutter_explist; /* the parser places the list of expressions here */

extern unsigned long _stutter__rest_sdbm; /* sdbm of "_REST" */
extern unsigned long _stutter__rest_id; /* unique symbol id of _REST symbol */

extern unsigned long _stutter_dispatch_info_sdbm; /* sdbm of _DISPATCH-INFO symbol */
extern unsigned long _stutter_dispatch_info_id; /* unique symbol id of _DISPATCH-INFO symbol */

extern unsigned long _stutter_lambda_sdbm; /* sdbm of LAMBDA symbol */
extern unsigned long _stutter_lambda_id; /* unique symbol id of LAMBDA symbol */

/* basic error codes */
enum {
	SE_UNDEF,	/* undefined symbol */
	SE_WTF,		/* cannot evaluate object */
	SE_PARM,	/* invalid parameters */
	SE_WFUNC,	/* invalid function */
	SE_TYPE,	/* type mismatch */
	SE_IPLIST,	/* non-symbol in parameter list of function */
	SE_PNUM,	/* insufficient parameters */
	SE_IO,		/* I/O error */
	SE_PARSE,	/* Parse error */
	SE_NFND,	/* Not found */
	SE_LIMIT,	/* Recursion limit exceeded */
	SE_MEM,		/* Memory allocation error */
	SE_MATH,	/* Math exception */
	SE_RET,		/* Return value (used by loop) */
	SE_COMMA,	/* Comma is illegal outside of backquote */
	SE_COMMA_SPLICE,	/* Comma-splice is illegal outside of backquote */
    SE_STOPREAD, /* Stop reading */
	SE_COMPILE,	/* error during compilation */

#ifdef WITH_SQLITE
	SE_SQLITE,	/* sqlite error */
#endif

	SE_LAST
};

extern char *errornames[]; /* defined in eval.c */

/* compilation: bound symbol */

typedef struct csym csym;

struct csym {
	int id;		/* symbol id */
	int stack;	/* position on stack (index) */
	csym *next;
};

/* compilation context */
typedef struct {
	csym *list;	/* list of bound symbols */
} compx;

/* create a compilation context */
compx * compx_create();						

/* destroy a compilation context */
void compx_free(compx *cx);					

/* push a bound symbol element on a compilation context */
void compx_pushcsym(compx *cx, int id, int stack);		

/* pop bound symbol element from a compilation context */
void compx_popcsym(compx *cx);						

#define eval(ctx, obj) eval_((ctx), (obj), 0)

/* evaluate expression
 * ctx is the variable context
 * obj is the object to be evaluated
 * tail is a boolean: whether the evaled object is in a tail position */
s_obj *eval_(varctx *ctx, s_obj *obj, const int tail);

/* user function executor
 * obj is the s-expression the call originates at
 * the tail flag is only used when evaluating macro results
 */
s_obj * funcexec(varctx *ctx, s_obj *obj, s_obj *func, s_obj *args, const int tail, int inhibit_macro_expansion);

/* load and evaluate from fp */
s_obj *evalfp(varctx *ctx, FILE *fp);

/* load and evaluate file */
s_obj *evalfile(varctx *ctx, const char *name);

/* reset parser (line count, parsed expresion list) */
void resetparser(void);

/* compile an expression */
s_obj *compile(varctx *ctx, s_obj *expr);

#endif
