#include <math.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include "builtins.h"
#include "eval.h"
#include "types.h"
#include "memory.h"

s_obj *sb_math_d2d(varctx *ctx, s_obj *parm, void *data, const int tail)
{
	s_obj *obj;
	double (* FUNC) (double);
	if(!nextarg(ctx, &obj, &parm))
		return create_error(SE_PNUM, "Math function needs a parameter");
	if(obj->type == ST_ERROR) return obj;
	if(obj->type != ST_NUM)
		return create_error(SE_TYPE, "Type mismatch");

	FUNC = data;
	return create_num(FUNC(obj->d.num.value));
}

void register_math_builtins(varctx *ctx)
{
	char *tou; char *c;
#define REG_D2D_MATH(FUNC) \
	tou = strdup(# FUNC); for(c=tou;*c;c++) *c = toupper(*c); \
	varctx_set(ctx, tou, create_builtin_with_data(sb_math_d2d, FUNC)); \
	free(tou);

	REG_D2D_MATH(acos)
	REG_D2D_MATH(asin)
	REG_D2D_MATH(atan)
	REG_D2D_MATH(ceil)
	REG_D2D_MATH(cos)
	REG_D2D_MATH(cosh)
	REG_D2D_MATH(fabs)
	REG_D2D_MATH(floor)
	REG_D2D_MATH(log)
	REG_D2D_MATH(log10)
	REG_D2D_MATH(sin)
	REG_D2D_MATH(sinh)
	REG_D2D_MATH(sqrt)
	REG_D2D_MATH(tan)
	REG_D2D_MATH(tanh)
}

/* XXX:

  double atan2(double y, double x) -- Compute arc tangent of y/x.
  double exp(double x -- Compute exponential of x
  double fmod(double x, double y) -- Divide x by y with integral quotient and return remainder.
  double frexp(double x, int *expptr) -- Breaks down x into mantissa and exponent of no.
  double ldexp(double x, int exp) -- Reconstructs x out of mantissa and exponent of two.
  double modf(double x, double *intptr) -- Breaks x into fractional and integer parts.
  double pow (double x, double y) -- Compute x raised to the power y.
  double tan(double x) -- Compute tangent of angle in radians.
  double tanh(double x) -- Compute the hyperbolic tangent of x.
*/

