#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <time.h>
#include "memory.h"
#include "eval.h"
#ifdef MINGW32
#include <winsock2.h>
#endif

#include "types.h"

#define GC_GEN2_PERIOD 2 /* generation 2 is checked for each Nth check of generation 1 */
#define GC_GEN3_PERIOD 4 /* generation 3 is checked for each Nth check of generation 2 */

int gc_gen2_count = GC_GEN2_PERIOD;
int gc_gen3_count = GC_GEN3_PERIOD;

unsigned long int _stutter_gc_mark = 1;
unsigned long int _stutter_gc_allocd = 0;
unsigned long int _stutter_gc_live = 0;
unsigned __int64 _stutter_varctx_id = 1;
unsigned long _stutter_symbol_id = 1;

varctx * symbol_name_hash = NULL;
varctx * class_hash = NULL;

int GC_ALLOC_LIMIT = 8192;
int GC_VERBOSE = 0;

/* list of allocated objects (generation 1) */
s_obj ** _stutter_gc_list;
int _stutter_gc_list_cnt;
size_t _stutter_gc_list_size;
s_obj ** _stutter_gc_list_tail;

/* list of allocated objects (generation 2) */
s_obj ** _stutter_gc_gen2;
int _stutter_gc_gen2_cnt;
size_t _stutter_gc_gen2_size;
s_obj ** _stutter_gc_gen2_tail;

/* list of allocated objects (generation 3) */
s_obj ** _stutter_gc_gen3;
int _stutter_gc_gen3_cnt;
size_t _stutter_gc_gen3_size;
s_obj ** _stutter_gc_gen3_tail;

/* list of protected objects */
s_obj ** _stutter_gc_prot;
int _stutter_gc_prot_cnt;
size_t _stutter_gc_prot_size;
s_obj ** _stutter_gc_prot_tail;

/* list of miscellaneous pointers to free at shutdown */
s_obj ** _stutter_miscfree;
int _stutter_miscfree_cnt;
size_t _stutter_miscfree_size;
s_obj ** _stutter_miscfree_tail;

/* recycle pool */
s_obj ** _stutter_recycle;
int _stutter_recycle_cnt;
size_t _stutter_recycle_size;
s_obj ** _stutter_recycle_tail;

#define GROWLIST_INIT(x) \
	x = (s_obj **) xmalloc(256 * sizeof(s_obj *)); \
	x ## _size = 256; \
	x ## _tail = x; \
	x ## _cnt = 0;

#define GROWLIST_PAD(x) \
	if(x ## _cnt == x ## _size) { \
		x ## _size = x ## _size * 2; \
		((x) = (s_obj **) xrealloc(x, x ## _size * sizeof(s_obj *))); \
		x ## _tail = x + x ## _cnt; \
	}

#define GROWLIST_ADD(x, o) \
	GROWLIST_PAD(x) \
	*(x ## _tail ++) = (o); \
	x ## _cnt ++;


#define GROWLIST_FREE(x) \
	free(x);

s_obj * class_precedence = NULL;

/* wrapper for calloc with error handling */
inline void *xcalloc(size_t nmemb, size_t size) {
	void *ptr = calloc(nmemb, size);
	if(!ptr) {
		perror("xcalloc");
		exit(EXIT_MEMORY_FAIL);
	}
	return ptr;
}

/* wrapper for malloc with error handling */
inline void *xmalloc(size_t size) {
	void *ptr = malloc(size);
	if(!ptr) {
		perror("xmalloc");
		exit(EXIT_MEMORY_FAIL);
	}
	return ptr;
}

/* wrapper for realloc with error handling */
inline void *xrealloc(void *ptr, size_t size) {
	ptr = realloc(ptr, size);
	if(!ptr) {
		perror("xrealloc");
		exit(EXIT_MEMORY_FAIL);
	}
	return ptr;
}

/* add object to recycle list (and set its serial to 0) */
inline void obj_recycle(s_obj *o)
{
	o->serial = 0;
	GROWLIST_ADD(_stutter_recycle, o);
}

/* try to take an element of the recycle list */
s_obj * obj_respawn()
{
	if(_stutter_recycle_tail > _stutter_recycle) {
		_stutter_recycle_tail --;
		_stutter_recycle_cnt --;

		memset(*_stutter_recycle_tail, 0, sizeof(s_obj));
		return *_stutter_recycle_tail;
	}

	return NULL;
}

/************************************** GARBAGE COLLECTOR ***/

/* list of protected roots */
s_protroot *gc_roots = NULL;

/* allocate s_protroot structure */
s_protroot *gc_alloc_protroot(varctx *ctx)
{
	s_protroot *p;
	p = (s_protroot *) xmalloc(sizeof(s_protroot));
	p->next = NULL;
	p->root = ctx;
	return p;
}

/* add an object to the root set */
void gc_protect_obj(s_obj *obj)
{
	gc_protect_root((varctx *) ((ptrint)obj | 1));
}

/* remove an object from the root set */
void gc_unprotect_obj(s_obj *obj)
{
	gc_unprotect_root((varctx *) ((ptrint)obj | 1));
}

/* protect a varctx root (scan for variables in it) */
void gc_protect_root(varctx *ctx)
{
	s_protroot *p;

	if(!gc_roots) {
		gc_roots = gc_alloc_protroot(ctx);
	} else {
		p = gc_alloc_protroot(ctx);
		p->next = gc_roots;
		gc_roots = p;
	}
}
/* unprotect a varctx root (no longer scan for variables in it) */
void gc_unprotect_root(varctx *ctx)
{
	s_protroot *p, *q;
	s_protroot head;
	head.root = NULL;
	head.next = gc_roots;

	for(p = &head; p->next != NULL; ) {
		if(p->next->root == ctx) {
			q = p->next;
			p->next = p->next->next;
			free(q);
			continue;
		}
		p = p->next;
	}
	gc_roots = head.next;
}

/* get current mark in protection stack */
s_mark gc_prot_mark(void) {
	return _stutter_gc_prot_cnt;
}

/* free up to mark in protection stack, except for object ret */
void gc_prot_free(s_mark mark, s_obj *ret) {
	if(ret) {
		GROWLIST_PAD(_stutter_gc_prot)
		*(_stutter_gc_prot + mark) = ret;
		_stutter_gc_prot_cnt = mark + 1;
		_stutter_gc_prot_tail = _stutter_gc_prot + _stutter_gc_prot_cnt;
	} else {
		_stutter_gc_prot_cnt = mark;
		_stutter_gc_prot_tail = _stutter_gc_prot + _stutter_gc_prot_cnt;
	}
}

/* add object to protection list */
inline void gc_protlist_add(s_obj *obj)
{
	GROWLIST_ADD(_stutter_gc_prot, obj);
}

/* record allocated object for GC */
void gc_record(s_obj *obj)
{
	GROWLIST_ADD(_stutter_gc_list, obj);
	_stutter_gc_allocd ++;

	gc_protlist_add(obj);
}

/* remove last allocated object from GC records (in case you want to free it yourself) */
void gc_unrecord()
{
	_stutter_gc_list_cnt --;
	_stutter_gc_list_tail --;
}

/* mark misc. pointer for free() */
void mark_miscfree(void *ptr)
{
	GROWLIST_ADD(_stutter_miscfree, (s_obj *) ptr);
}

/* misc pointers in miscfree list */
void free_miscfree()
{
	s_obj **n;
	for(n = _stutter_miscfree; n != _stutter_miscfree_tail; n++)
		free(*n);
	_stutter_miscfree_tail = _stutter_miscfree;
	_stutter_miscfree_cnt = 0;
}

/* checks whether the link should be followed
 * (not if the target object has the link in its weakptr list */
inline void gc_follow(s_obj **obj)
{
	weakptr *wp;
	if(!*obj) return;
	for(wp = (*obj)->weak; wp; wp=wp->next) {
//		fprintf(stderr, "follow? *obj=%p obj=%p ptr=%p *ptr=%p\n", *obj, obj, wp->ptr, *wp->ptr);
		if(wp->ptr == obj) {
//			fprintf(stderr, "  no.\n");
			return;
		}
 	}

	gc_mark_obj(*obj);
}

/* mark variables reachable from object */
void gc_mark_obj(s_obj *obj)
{
	int i, ds;
	s_obj **a;

	if(!obj) return;
	if(obj->mark == _stutter_gc_mark)
		return; /* why hello, we looked here already */

	obj->mark = _stutter_gc_mark;

	switch(obj->type) {
		case ST_CONS:
			gc_follow(&obj->d.cons.car);
			gc_follow(&obj->d.cons.cdr);
			break;
		case ST_VARCTX:
			gc_mark(obj->d.varctx.ctx);
			break;
		case ST_FUNC:
			gc_mark_obj(obj->d.func.parm);
			gc_mark_obj(obj->d.func.exp);
			gc_mark_obj(obj->d.func.docstr);
			gc_mark_obj(obj->d.func.closure);
			break;
		case ST_DICT:
			for(i=0; i<obj->d.dict.size; i++) {
				if(obj->d.dict.data[i].name != NULL) {
					gc_mark_obj(obj->d.dict.data[i].name);
					gc_follow(&obj->d.dict.data[i].value);
				}
			}
			break;
		case ST_STREAM:
			switch(obj->d.stream.type) {
				case STREAM_FUNC:
					gc_mark_obj(obj->d.stream.s.func);
					break;
				default:
					;
			}
			break;
		case ST_ERROR:
			if(obj->d.error.obj) {
				gc_mark_obj(obj->d.error.obj);
			}
			if(obj->d.error.trace) {
				gc_mark_obj(obj->d.error.trace);
			}
			break;
		case ST_ARRAY:
			ds = 1; /* compute array size */
			for(i=0; i<obj->d.array.dim; i++)
				ds *= obj->d.array.size[i];
			a = obj->d.array.data;
			for(i=0;i<ds;i++) {
				gc_follow(a);
				a++;
			}
			break;

		case ST_CLASS:
			gc_mark_obj(obj->d.clos_class.parents);
			gc_mark_obj(obj->d.clos_class.children);
			gc_mark_obj(obj->d.clos_class.slots);
			gc_mark_obj(obj->d.clos_class.prec);
			gc_mark_obj(obj->d.clos_class.name);
			gc_mark_obj(obj->d.clos_class.ancestry);

			for(i=0; i<obj->d.clos_class.allslotsnum; i++) {
				gc_mark_obj(obj->d.clos_class.allslots[i].stor);
				gc_mark_obj(obj->d.clos_class.allslots[i].slot);
			}

			for(i=0; i<obj->d.clos_class.data_size; i++)
				gc_follow(&obj->d.clos_class.data[i]);

			gc_mark(obj->d.clos_class.initarg);
			gc_mark(obj->d.clos_class.slotname);

			break;

		case ST_SLOT:
			gc_mark_obj(obj->d.clos_slot.initform);
			gc_mark_obj(obj->d.clos_slot.name);
			gc_mark_obj(obj->d.clos_slot.accessor);
			gc_mark_obj(obj->d.clos_slot.writer);
			gc_mark_obj(obj->d.clos_slot.reader);
			gc_mark_obj(obj->d.clos_slot.initarg);
			break;

		case ST_OBJECT:
			gc_mark_obj(obj->d.clos_object.type);
			for(i=0; i<obj->d.clos_object.data_size; i++)
				gc_follow(&obj->d.clos_object.data[i]);
			break;

		case ST_GENERIC:
			gc_mark_obj(obj->d.clos_generic.m_primary);
			gc_mark_obj(obj->d.clos_generic.m_before);
			gc_mark_obj(obj->d.clos_generic.m_after);
			gc_mark_obj(obj->d.clos_generic.m_around);

			for(i=0; i<obj->d.clos_generic.dispatch_size; i++) {
				gc_mark_obj(obj->d.clos_generic.dispatch[i].method);

				dispatch_signature *sgnt;
				for(sgnt = obj->d.clos_generic.dispatch[i].signature; sgnt && sgnt->type != -1 ; sgnt++)
					gc_mark_obj(sgnt->value);

			}
			break;

		case ST_METHOD:
			for(i=0; i<obj->d.clos_method.argn; i++) {
				gc_mark_obj(obj->d.clos_method.args[i].name);
				gc_mark_obj(obj->d.clos_method.args[i].value);
			}

			gc_mark_obj(obj->d.clos_method.miscargs);
			gc_mark_obj(obj->d.clos_method.code);
			gc_mark_obj(obj->d.clos_method.next);
			break;

		case ST_DISPATCH:
			gc_mark_obj(obj->d.clos_dispatch.generic);
			gc_mark_obj(obj->d.clos_dispatch.method);
			i=0;
			while(obj->d.clos_dispatch.argeval[i]) {
				gc_mark_obj(obj->d.clos_dispatch.argeval[i++]);
			}
			break;

		default:
			;
	}
}

/* mark variables in context */
void gc_mark(varctx *ctx)
{
	int i;
	s_var *var;

	gc_mark_obj(ctx->ref);
	if(ctx->parent) gc_mark(ctx->parent);

	if(ctx->mark == _stutter_gc_mark)
		return; /* why hello, we looked here already */
	ctx->mark = _stutter_gc_mark;

	for(i=0; i<ctx->ns; i++) {
		for(var = ctx->hash[i]; var; var = var->next) {
			if(var->value) {
				gc_follow(&var->value);
			}
		}
	}
}

/* free all data associated with object */
inline void gc_harvest(s_obj *obj)
{
	weakptr *ptr, *n;

	/* set all weak pointers to nil, but only if they still point here */
	for(ptr = obj->weak; ptr;) {
//        fprintf(stderr, "harvest: %d == %d && %p == %p\n", (int)ptr->host->serial, (int)ptr->serial, *ptr->ptr, obj);
		if(ptr->host->serial == ptr->serial && /* host object still exists */
			*ptr->ptr == obj) {

			*ptr->ptr = create_nil();
		}

		n = ptr->next;
		free(ptr);
		ptr = n;
	}

	obj->weak = NULL;

	freedata(obj);
	obj_recycle(obj);
}

/* do garbage collection. returns number of objects freed */
int gc_sweep(varctx *ctx)
{
	s_obj **a;
	s_protroot *p;
	int sw = 0;

	/* mark all reachable objects */
	_stutter_gc_mark ++;
	if(ctx) gc_mark(ctx);

	for(p = gc_roots; p; p=p->next)
		if((ptrint) (p->root) & 1)
			gc_mark_obj((s_obj *) ((ptrint) (p->root) ^ 1));
		else
			gc_mark(p->root);

	for(a = _stutter_gc_prot; a != _stutter_gc_prot_tail; a++) {
		gc_mark_obj(*a);
	}

	/* well, all classes referenced in type db should be
	 * registered in class_hash anyway */
//	for(int i = 0; i < next_typeid; i++)
//		gc_mark_obj(typedb[i].cl);

	/* sweep generation 1 */
	for(a = _stutter_gc_list; a != _stutter_gc_list_tail; ) {
		if((*a)->mark < _stutter_gc_mark) {

			sw++;
			gc_harvest(*a);
			_stutter_gc_list_tail --;
			*a = *_stutter_gc_list_tail;
			_stutter_gc_list_cnt --;
		} else {

			/* move object to generation 2 */

			GROWLIST_ADD(_stutter_gc_gen2, *a);

			_stutter_gc_list_tail --;
			*a = *_stutter_gc_list_tail;
			_stutter_gc_list_cnt --;
		}
	}

	gc_gen2_count --;
	if(gc_gen2_count == 0) {
		gc_gen2_count = GC_GEN2_PERIOD;
		gc_gen3_count --;

		for(a = _stutter_gc_gen2; a != _stutter_gc_gen2_tail; ) {
			if((*a)->mark < _stutter_gc_mark) {

				sw++;
				gc_harvest(*a);
				_stutter_gc_gen2_tail --;
				*a = *_stutter_gc_gen2_tail;
				_stutter_gc_gen2_cnt --;
			} else {

				/* move object to generation 3 */

				GROWLIST_ADD(_stutter_gc_gen3, *a);

				_stutter_gc_gen2_tail --;
				*a = *_stutter_gc_gen2_tail;
				_stutter_gc_gen2_cnt --;
			}
		}

		if(gc_gen3_count == 0) {
			gc_gen3_count = GC_GEN3_PERIOD;

			for(a = _stutter_gc_gen3; a != _stutter_gc_gen3_tail; ) {
				if((*a)->mark < _stutter_gc_mark) {

					sw ++;
					gc_harvest(*a);
					_stutter_gc_gen3_tail --;
					*a = *_stutter_gc_gen3_tail;
					_stutter_gc_gen3_cnt--;
				} else a++;
			}
		}
	}

	_stutter_gc_live += (_stutter_gc_allocd - sw);

	if(GC_VERBOSE) {
		fprintf(stderr, "GC: %d freed, %ld estimated live (%d %d %d), pool=%d\n", sw, _stutter_gc_live, _stutter_gc_list_cnt, _stutter_gc_gen2_cnt, _stutter_gc_gen3_cnt, _stutter_recycle_cnt);
	}

	_stutter_gc_allocd = 0;

	return sw;
}

/**************************************** VARIABLE CONTEXTS ***/

/* create variable context */
varctx *varctx_create(varctx *parent, const int size)
{
	varctx *vc;

	assert(!((size - 1) & size));

	vc = (varctx *) xmalloc(sizeof(varctx));
	vc->parent = parent;
	vc->ns = size;
	vc->nsm = size - 1;
	vc->id = _stutter_varctx_id++;
	vc->weak = 0;
	vc->ref = NULL;
	vc->mark = 0;

	if(size > 0) {
		vc->defer = 0;
		vc->hash = (s_var **) xcalloc(size, sizeof(s_var *));
	} else {
		vc->defer = 1;
	}

	return vc;
}

s_var *alloc_s_var(void)
{
	s_var *s;
	s = (s_var *) xcalloc(1, sizeof(s_var));
	return s;
}

/* store a variable inside a ctx */
void varctx_store_s(varctx *ctx, unsigned long sdbmv, unsigned long id, s_obj *val)
{
	s_var *sv;
	s_var **sf;

	sv = alloc_s_var();
	sv->name = NULL;
	sv->sdbm = sdbmv;
	sv->id = id;
	sv->value = val;

	sf = &(ctx->hash[sdbmv & ctx->nsm]);
	while(*sf) sf = &((*sf)->next);
	*sf = sv;
}

/* store a variable inside a ctx */
void varctx_store(varctx *ctx, char *name, s_obj *val)
{
	unsigned long sdbmv = sdbm(name);
	varctx_store_s(ctx, sdbmv, get_symbolid(name, sdbmv), val);
}

/* store a variable inside a ctx - with given sdbm value */
void varctx_store_sdbm(varctx *ctx, char *name, s_obj *val, unsigned long sdbmv)
{
	varctx_store_s(ctx, sdbmv, get_symbolid(name, sdbmv), val);
}

/* empty a varctx - erase all lists */
void varctx_empty(varctx *ctx)
{
	int i;
	s_var *v, *f;

	for(i=0; i<ctx->ns; i++) {
		for(v = ctx->hash[i]; v;) {
//			weak_unref(&v->value);
			f = v;
			v = v->next;
			free(f->name);
			free(f);
		}
		ctx->hash[i] = NULL;
	}
}

/* get symbol name by symbol id */
char * get_symbolname(unsigned long sdbmv, unsigned long id)
{
	s_var *sv;

	sv = symbol_name_hash->hash[sdbmv & symbol_name_hash->nsm];
	while(sv) {
		if(sv->id == id)
			return sv->name;
		sv = sv->next;
	}

	return NULL;
}

/* get unique symbol id of a symbol */
unsigned long get_symbolid(char *name, unsigned long sdbmv)
{
	s_var *sv, **sf;

	if(!symbol_name_hash)
		symbol_name_hash = varctx_create(NULL, SYMBOL_NAME_HASH_SIZE);

	sv = symbol_name_hash->hash[sdbmv & symbol_name_hash->nsm];
	while(sv) {
		if(!strcmp(name, sv->name))
			return sv->id;
		sv = sv->next;
	}

	sv = alloc_s_var();
	sv->name = strdup(name);
	sv->sdbm = sdbmv;
	sv->id = _stutter_symbol_id ++;

	sf = &(symbol_name_hash->hash[sdbmv & symbol_name_hash->nsm]);
	while(*sf) sf = &((*sf)->next);
	*sf = sv;

	return sv->id;
}

/* utility function: get symbol id by string */
unsigned long symbolid(char *name)
{
	return get_symbolid(name, sdbm(name));
}

/* get value of a variable from a ctx (local only)
 * returns NULL if variable doesn't exist */
s_var *varctx_retr_s(varctx *ctx, unsigned long sdbmv, unsigned long id)
{
	s_var *sv;
	sv = ctx->hash[sdbmv & ctx->nsm];
	while(sv) {
		if(id == sv->id)
			return sv;
		sv = sv->next;
	}

	return NULL;
}

/* get value of a variable from a ctx (local only)
 * returns NULL if variable doesn't exist */
s_var *varctx_retr(varctx *ctx, char *name)
{
	unsigned long sdbmv = sdbm(name);
	return varctx_retr_s(ctx, sdbmv, get_symbolid(name, sdbmv));
}

/* get value of a variable from a ctx (local only)
 * returns NULL if variable doesn't exist */
s_var *varctx_retr_sdbm(varctx *ctx, char *name, unsigned long sdbmv)
{
	return varctx_retr_s(ctx, sdbmv, get_symbolid(name, sdbmv));
}

/* get value of a variable from a ctx (with parent fallback)
 * returns NULL if variable doesn't exist */
s_var *varctx_get_s(varctx *ctx, unsigned long sdbmv, unsigned long id)
{
	s_var *v;
	v = varctx_retr_s(ctx, sdbmv, id);
	if(!v && ctx->parent) v = varctx_get_s(ctx->parent, sdbmv, id);
	return v;
}

/* get value of a variable from a ctx (with parent fallback)
 * returns NULL if variable doesn't exist */
s_var *varctx_get_sdbm(varctx *ctx, char *name, unsigned long sdbmv)
{
	return varctx_get_s(ctx, sdbmv, get_symbolid(name, sdbmv));
}

/* get value of a variable from a ctx (with parent fallback)
 * returns NULL if variable doesn't exist */
s_var *varctx_get(varctx *ctx, char *name)
{
	return varctx_get_sdbm(ctx, name, sdbm(name));
}

/* set/store a variable inside a ctx
 * (updates if the var already exists) */
void varctx_set_s(varctx *ctx, unsigned long sdbmv, unsigned long id, s_obj *val)
{
	s_var *v;
	if((v = varctx_retr_s(ctx, sdbmv, id))) {
		v->value = val;
		return;
	}

	if(ctx->weak && ctx->parent) varctx_set_s(ctx->parent, sdbmv, id, val);
	else varctx_store_s(ctx, sdbmv, id, val);
}

/* set/store a variable inside a ctx
 * (updates if the var already exists) */
void varctx_set_sdbm(varctx *ctx, char *name, s_obj *val, unsigned long sdbmv)
{
	s_var *v;
	if((v = varctx_retr_sdbm(ctx, name, sdbmv))) {
		v->value = val;
		return;
	}

	if(ctx->weak && ctx->parent) varctx_set_sdbm(ctx->parent, name, val, sdbmv);
	else varctx_store_sdbm(ctx, name, val, sdbmv);
}

/* set/store a variable inside a ctx
 * (updates if the var already exists) */
void varctx_set(varctx *ctx, char *name, s_obj *val)
{
	s_var *v;
	unsigned long sdbmv = sdbm(name);
	unsigned long id = get_symbolid(name, sdbmv);

	if((v = varctx_retr_s(ctx, sdbmv, id))) {
		v->value = val;
		return;
	}

	if(val->type == ST_BUILTIN) {
		if(!val->d.builtin.symbol) {
			val->d.builtin.symbol = get_symbolname(sdbmv, id);
			val->d.builtin.symbolid = id;
		}
	}

	varctx_store_s(ctx, sdbmv, id, val);
}

/* compute hash value for a string */
unsigned long sdbm(char *str)
{
	unsigned long hash = 0;
	int c;

	while((c = *str++))
		hash = c + (hash << 6) + (hash << 16) - hash;

	return hash;
}

/* free varctx */
void varctx_free(varctx *ctx)
{
	gc_unprotect_root(ctx);

	if(!ctx->defer) {
		varctx_empty(ctx);
		free(ctx->hash);
	}

	free(ctx);
}

/* dump contents of varctx to stderr */
s_obj * varctx_dump(varctx *ctx)
{
	int i;
	s_var *var;
	s_obj *o;

	o = stream_printf(ctx, "---------- %p\n", ctx);
	if(o->type == ST_ERROR) return o;

	while(ctx->parent) {
		for(i=0; i<ctx->ns; i++) {
			for(var = ctx->hash[i]; var; var = var->next) {
				o = stream_printf(ctx, "%p %s = ", var->value, get_symbolname(var->sdbm, var->id));
				if(o->type == ST_ERROR) return o;
				o = printobj(ctx, var->value);
				if(o->type == ST_ERROR) return o;
				o = stream_stdout_write(ctx, '\n');
				if(o->type == ST_ERROR) return o;
			}
		}

		o = stream_printf(ctx, "========== %p\n", ctx->parent);
		if(o->type == ST_ERROR) return o;
		ctx = ctx->parent;
	}
	o = stream_printf(ctx, "---------------------------\n");
	if(o->type == ST_ERROR) return o;
	return create_nil();
}

/****************************************** INIT ***/

/* initialize Stutter */
void stutter_init()
{
	int i;

#ifdef MINGW32
	WORD wVersionRequested = MAKEWORD(2, 2);
	WSADATA wsaData;
	WSAStartup(wVersionRequested, &wsaData);
#endif

	_stutter_gc_allocd = 0;
	_stutter_gc_live = 0;
	_stutter_varctx_id = 1;
	_stutter_symbol_id = 1;

	symbol_name_hash = NULL;	
	_stutter__rest_sdbm = sdbm("_REST");
	_stutter__rest_id = get_symbolid("_REST", _stutter__rest_sdbm);

	_stutter_dispatch_info_sdbm = sdbm("_DISPATCH-INFO");
	_stutter_dispatch_info_id = get_symbolid("_DISPATCH-INFO", _stutter_dispatch_info_sdbm);

	_stutter_lambda_sdbm = sdbm("LAMBDA");
	_stutter_lambda_id = get_symbolid("LAMBDA", _stutter_lambda_sdbm);

	srand(time(NULL));

	GROWLIST_INIT(_stutter_gc_prot);
	GROWLIST_INIT(_stutter_gc_list);
	GROWLIST_INIT(_stutter_gc_gen2);
	GROWLIST_INIT(_stutter_gc_gen3);
	GROWLIST_INIT(_stutter_miscfree);
	GROWLIST_INIT(_stutter_recycle);

	objcache_init();

	class_hash = varctx_create(NULL, 64);
	gc_protect_root(class_hash);

	next_typeid = 0;
	typedb = NULL;
	for(i=0; i<TYPE_LAST_MARK; i++)
		gettypeid();
	class_precedence = NULL;
}

/* shut down Stutter */
void stutter_shutdown()
{
	s_obj **a;

	gc_unprotect_root(class_hash);
	varctx_free(symbol_name_hash);
	gc_prot_free(0, NULL);
	gc_gen2_count = 1;
	gc_gen3_count = 1;
	gc_sweep(NULL);
	varctx_free(class_hash);
	free_miscfree();
	free(typedb);

	for(a = _stutter_recycle; a != _stutter_recycle_tail; a++) {
		free(*a);
	}

	GROWLIST_FREE(_stutter_gc_prot);
	GROWLIST_FREE(_stutter_gc_list);
	GROWLIST_FREE(_stutter_gc_gen2);
	GROWLIST_FREE(_stutter_gc_gen3);
	GROWLIST_FREE(_stutter_miscfree);
	GROWLIST_FREE(_stutter_recycle);
}

/********************************** newhash */

typedef  unsigned long int  u4;   /* unsigned 4-byte type */
typedef  unsigned     char  u1;   /* unsigned 1-byte type */

/* The mixing step */
#define mix(a,b,c) \
{ \
	a=a-b;  a=a-c;  a=a^(c>>13); \
	b=b-c;  b=b-a;  b=b^(a<<8);  \
	c=c-a;  c=c-b;  c=c^(b>>13); \
	a=a-b;  a=a-c;  a=a^(c>>12); \
	b=b-c;  b=b-a;  b=b^(a<<16); \
	c=c-a;  c=c-b;  c=c^(b>>5);  \
	a=a-b;  a=a-c;  a=a^(c>>3);  \
	b=b-c;  b=b-a;  b=b^(a<<10); \
	c=c-a;  c=c-b;  c=c^(b>>15); \
}

/* The whole new hash function */
u4 hash(k, length, initval)
	register u1 *k;        /* the key */
	u4           length;   /* the length of the key in bytes */
	u4           initval;  /* the previous hash, or an arbitrary value */
{
	register u4 a,b,c;  /* the internal state */
	u4          len;    /* how many key bytes still need mixing */

	/* Set up the internal state */
	len = length;
	a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
	c = initval;         /* variable initialization of internal state */

	/*---------------------------------------- handle most of the key */
	while (len >= 12)
	{
		a=a+(k[0]+((u4)k[1]<<8)+((u4)k[2]<<16) +((u4)k[3]<<24));
		b=b+(k[4]+((u4)k[5]<<8)+((u4)k[6]<<16) +((u4)k[7]<<24));
		c=c+(k[8]+((u4)k[9]<<8)+((u4)k[10]<<16)+((u4)k[11]<<24));
		mix(a,b,c);
		k = k+12; len = len-12;
	}

	/*------------------------------------- handle the last 11 bytes */
	c = c+length;
	switch(len)              /* all the case statements fall through */
	{
		case 11: c=c+((u4)k[10]<<24);
		case 10: c=c+((u4)k[9]<<16);
		case 9 : c=c+((u4)k[8]<<8);
				 /* the first byte of c is reserved for the length */
		case 8 : b=b+((u4)k[7]<<24);
		case 7 : b=b+((u4)k[6]<<16);
		case 6 : b=b+((u4)k[5]<<8);
		case 5 : b=b+k[4];
		case 4 : a=a+((u4)k[3]<<24);
		case 3 : a=a+((u4)k[2]<<16);
		case 2 : a=a+((u4)k[1]<<8);
		case 1 : a=a+k[0];
				 /* case 0: nothing left to add */
	}
	mix(a,b,c);
	/*-------------------------------------------- report the result */
	return c;
}

