#include <stdlib.h>
#include "stutter.h"

/* builtin: quote */
BUILTIN(sb_quote)
{
	if(parm->type != ST_CONS)
		return create_error(SE_PARM, "QUOTE needs a parameter");

	return parm->d.cons.car;
}

/* quote an object for passing.. returns (quote obj) */
s_obj *quoteobj(s_obj *obj)
{
	s_obj *b = create_sform(sb_quote);
	b->d.builtin.printprefix = '\'';
	return create_cons(b, create_cons(obj, create_nil()));
}

/* builtin: lambda */
BUILTIN(sb_lambda)
{
	s_obj *prm, *exp = NULL, *first, *cur;
	s_obj *docstr = NULL;

	if(!nextparm(&prm, &parm))
		return create_error(SE_PNUM, "Insufficient parameters for lambda");
	if(!prm || (prm->type != ST_CONS && prm->type != ST_NIL))
		return create_error(SE_PARM, "First parameter of lambda must be a list of argument names");

	first = cur = NULL;
	while(nextparm(&exp, &parm)) {
		if(!cur) {
			if(!docstr && exp->type == ST_STRING) docstr = exp;
			else
				first = cur = create_cons(exp, create_nil());
		} else {
			cur->d.cons.cdr = create_cons(exp, create_nil());
			cur = cur->d.cons.cdr;
		}
	}

	if(!first && docstr) {
		first = create_cons(docstr, create_nil());
		docstr = NULL;
	}

	exp = first ? first : create_nil();
	cur = create_func(prm, exp);
	cur->d.func.docstr = docstr;
	cur->d.func.macro = (ptrint) data;
		/* 1 for macro, 2 for r-macro
		 * 4 for symbolic-macro, 5 for r-symbolic-macro */

	if(ctx->parent) {
		/* varctx isn't global.. create a closure */
		if(!ctx->ref) {
			cur->d.func.closure = create_obj_varctx(ctx);
			ctx->ref = cur->d.func.closure;
		} else cur->d.func.closure = ctx->ref;

		/* keep all parents */
		ctx = ctx->parent;
		while(ctx->parent) {
			if(!ctx->ref) ctx->ref = create_obj_varctx(ctx);
			ctx = ctx->parent;
		}
	}

	return cur;
}

/* builtin: if */
BUILTIN(sb_if)
{
	s_obj *obj;
	s_obj *val;
	s_obj *tr;

	if(!nextparm(&obj, &parm))
		return create_error(SE_PNUM, "Insufficient parameters for if");
	val = eval(ctx, obj);
	if(val->type == ST_ERROR) return val;

	if(!nextparm(&tr, &parm))
		return create_error(SE_PARM, "Insufficient parameters for if");

	if(val->type == ST_NIL) {
		/* false */
		if(!nextparm(&tr, &parm))
			return create_nil();
		else
			return eval_(ctx, tr, tail);
	} else {
		/* true */
		return eval_(ctx, tr, tail);
	}
}

/* builtin: subctx */
BUILTIN(sb_subctx)
{
	s_obj *obj, *j = NULL;
	varctx *fctx;

	fctx = varctx_create(ctx, 64);

	while(nextparm(&obj, &parm)) {
		j = eval(fctx, obj);
		if(j->type == ST_ERROR) {
			if(!fctx->ref) varctx_free(fctx);
			return j;
		}
	}

	if(!fctx->ref) varctx_free(fctx);

	return j ? j : create_nil();
}

/* builtin: do/let */
BUILTIN(sb_do)
{
	s_obj *bind, *test = NULL, *asg, *i, *obj, *var, *ret = NULL;
	s_mark mark;
	s_obj *vals[FUNC_MAXPARMS];
	varctx *nctx;
	int let = (ptrint) data;
	int nasg;

	if(!nextparm(&bind, &parm))
		return create_error(SE_PNUM, "Insufficient parameters for do/let");
	if(bind->type != ST_CONS && bind->type != ST_NIL)
		return create_error(SE_TYPE, "First argument of do/let has to be a list");
	if(!let) {
		if(!nextparm(&test, &parm))
			return create_error(SE_PNUM, "Insufficient parameters for do/let");
		if(test->type != ST_CONS)
			return create_error(SE_TYPE, "Second argument of do has to be a list");
	}

	nctx = varctx_create(ctx, FUNC_LOCALCTX_SIZE);
	nctx->weak = 1;

	/* assign initial values */

	i = bind; nasg = 0;
	while(nextparm(&asg, &i)) {
		if(nasg == FUNC_MAXPARMS) {
			if(!nctx->ref) varctx_free(nctx);
			return create_error(SE_PNUM, "Too many assignment clauses in do/let");
		}

		if(asg->type != ST_CONS) {
			if(!nctx->ref) varctx_free(nctx);
			return create_error(SE_TYPE, "First argument of do/let has to be a list of lists");
		}

		nextparm(&var, &asg);
		if(var->type != ST_SYMBOL) {
			if(!nctx->ref) varctx_free(nctx);
			return create_error(SE_TYPE, "Do/let can't assign to a non-symbol");
		}

		if(!nextarg(ctx, &obj, &asg)) {
			obj = create_nil();
		} else if(obj->type == ST_ERROR) {
			if(!nctx->ref) varctx_free(nctx);
			if(!let && obj->d.error.code == SE_RET) return obj->d.error.obj ? obj->d.error.obj : create_nil();
			return obj;
		}

		varctx_store_s(nctx, var->d.symbol.sdbm, var->d.symbol.id, obj);
		nasg ++;
	}

	/* the loop */

	while(1) {
		mark = gc_prot_mark();

		if(!let) { /* (do), test condition */
			i = test;
			nextparm(&ret, &i);
			ret = eval(nctx, ret);
			if(ret->type == ST_ERROR) {
				if(!nctx->ref) varctx_free(nctx);
				if(!let && ret->d.error.code == SE_RET) return ret->d.error.obj ? ret->d.error.obj : create_nil();
				return ret;
			}
			if(ret->type != ST_NIL) {
				if(nextarg(nctx, &obj, &i)) { /* return value specified */
					if(!nctx->ref) varctx_free(nctx);
					return obj;
				}

				if(!nctx->ref) varctx_free(nctx); /* return value not specified, returning test result */
				return ret;
			}
		}

		/* execute body */
		i = parm;
		while(i->type == ST_CONS) {
			ret = eval(nctx, i->d.cons.car);
			if(ret->type == ST_ERROR) {
				if(!nctx->ref) varctx_free(nctx);
				if(!let && ret->d.error.code == SE_RET) return ret->d.error.obj ? ret->d.error.obj : create_nil();
				return ret;
			}
			i = i->d.cons.cdr;
		}

		if(let) break; else {
			/* execute adjustments */

			i = bind;
			nasg = 0;
			while(nextparm(&asg, &i)) { /* evaluate values */
				vals[nasg] = varctx_retr_s(nctx, asg->d.cons.car->d.symbol.sdbm, asg->d.cons.car->d.symbol.id)->value;
				asg = asg->d.cons.cdr;
				if(asg->type != ST_CONS) {
					nasg++; continue;
				}

				asg = asg->d.cons.cdr;
				if(asg->type == ST_CONS) {
					asg = asg->d.cons.car;
					obj = eval(nctx, asg);
					if(obj->type == ST_ERROR) {
						if(!nctx->ref) varctx_free(nctx);
						if(!let && obj->d.error.code == SE_RET) return obj->d.error.obj ? obj->d.error.obj : create_nil();
						return obj;
					}
					vals[nasg] = obj;
				}
				nasg++;
			}

			/* assign values */
			nasg = 0; i = bind;
			while(nextparm(&asg, &i)) {
				varctx_set_s(nctx, asg->d.cons.car->d.symbol.sdbm, asg->d.cons.car->d.symbol.id, vals[nasg]);
				nasg ++;
			}

			gc_prot_free(mark, NULL);
		}
	}

	if(!nctx->ref) varctx_free(nctx);

	return ret ? ret : create_nil();
}

/* builtin: catch */
BUILTIN(sb_catch)
{
	s_obj *list, *car;
	s_obj *msgo;
	s_obj *obj = NULL;
	int cgt;

	if(!nextarg(ctx, &list, &parm))
		return create_error(SE_PNUM, "Insufficient parameters for catch");
	if(list->type == ST_ERROR) return list;
	if(list->type != ST_CONS && list->type != ST_T)
		return create_error(SE_TYPE, "Error codes parameter for catch must be a list or T");

	/* check list for types */
	if(list->type == ST_CONS) {
		for(car = list; car->type == ST_CONS; car = car->d.cons.cdr) {
			if(car->d.cons.car->type != ST_INT)
				return create_error(SE_TYPE, "Error codes must be integers");
		}
	}

	ARG_FUNCTION(func);

	while(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) {
			/* it's an error, look in the list whether we should catch it */
			cgt = 0;
			if(list->type == ST_T) {
				cgt = 1;
			} else {
				for(car = list; car->type == ST_CONS; car = car->d.cons.cdr) {
					if(car->d.cons.car->d.integer.value == obj->d.error.code) {
						cgt = 1;
						break;
					}
				}
			}

			if(cgt) {
				/* caught it, call handler */
				if(obj->d.error.msg) msgo = create_string(obj->d.error.msg); else msgo = create_nil();
				func = create_cons(func,
					create_cons(create_int(obj->d.error.code),
						create_cons(msgo,
							create_cons(obj->d.error.obj ? quoteobj(obj->d.error.obj) : create_nil(),
								create_nil()))));
				obj = eval(ctx, func);
				break;
			}

			return obj;
		}
	}

	return obj ? obj : create_nil();
}

/* builtin: defmethod */
BUILTIN(sb_defmethod)
{
	s_obj *name;
	s_obj *place, *args = NULL;
	s_obj *code;
	s_obj *method;
	s_obj *iter;
	s_obj *arg;
	int argn;

	int mplace = CL_METHOD_PRIMARY;

	unsigned long _BEFORE = symbolid("BEFORE");
	unsigned long _PRIMARY = symbolid("PRIMARY");
	unsigned long _AFTER = symbolid("AFTER");
	unsigned long _AROUND = symbolid("AROUND");
	unsigned long _PLUS = symbolid("PLUS");
	unsigned long _EQL = symbolid("EQL");

	/* grab method name */
	if(!nextparm(&name, &parm))
		return create_error(SE_PNUM, "Defmethod needs parameter: NAME");
	if(name->type != ST_SYMBOL)
		return create_error(SE_TYPE, "Defmethod: NAME is expected to be a symbol");

	/* next argument is either method place or the arglist */
	if(!nextparm(&place, &parm))
		return create_error(SE_PNUM, "Defmethod needs parameter: ARGLIST");
	if(place->type == ST_SYMBOL) {
		/* method place it is */

		if(place->d.symbol.id == _BEFORE)
			mplace = CL_METHOD_BEFORE;
		else if(place->d.symbol.id == _PRIMARY)
			mplace = CL_METHOD_PRIMARY;
		else if(place->d.symbol.id == _AFTER)
			mplace = CL_METHOD_AFTER;
		else if(place->d.symbol.id == _AROUND)
			mplace = CL_METHOD_AROUND;
		else
			return create_error(SE_PARM, "Defmethod: invalid method place");

		/* grab arglist after this */
		if(!nextparm(&args, &parm))
			return create_error(SE_PNUM, "Defmethod needs parameter: ARGLIST (after PLACE)");
		if(args->type != ST_CONS && args->type != ST_NIL)
			return create_error(SE_TYPE, "Defmethod: ARGLIST must be a list");

	} else if(place->type == ST_CONS || place->type == ST_NIL) {
		/* arglist it is */
		args = place;
	} else return create_error(SE_TYPE, "Defmethod: NAME must be followed by a PLACE or an ARGLIST");

	/* everything after this is code for the method */
	code = parm;

	method = create_method();
	method->d.clos_method.place = mplace;
	method->d.clos_method.code = code;

	/* count number of arguments */
	argn = 0;
	iter = args;
	while(nextparm(&arg, &iter)) {
		/* break after :+
		 * (the following arguments aren't considered for dispatch) */
		if(arg->type == ST_SYMBOL && arg->d.symbol.id == _PLUS && arg->d.symbol.colon) break;
		argn ++;
	}

	if(argn) {
		/* make place for the argument structure */
		method->d.clos_method.args = (method_arg *) xcalloc(argn, sizeof(method_arg));
		method->d.clos_method.argn = argn;

		/* parse arguments */
		argn = 0;
		iter = args;

		while(nextparm(&arg, &iter)) {
			if(arg->type == ST_SYMBOL) {
				/* end of significant arglist? */
				if(arg->d.symbol.id == _PLUS && arg->d.symbol.colon)
					break;

				/* simple untyped arg */
				method->d.clos_method.args[argn].name = arg;
				method->d.clos_method.args[argn].type = -1;
			} else if(arg->type == ST_CONS) {
				/* some kind of classified arg. */

				s_obj *iter2 = arg;
				s_obj *item;

				nextparm(&item, &iter2);
				if(item->type != ST_SYMBOL)
					return create_error(SE_TYPE, "Defmethod: argument name in classified pair must be a symbol");

				method->d.clos_method.args[argn].name = item;

				if(nextparm(&item, &iter2)) {
					/* classification */

					if(item->type == ST_SYMBOL) {
						/* should be a type */

						int typeid = symbol_to_typeid(item);
						if(typeid == -1)
							return create_error_with_obj(SE_PARM, "Defmethod: invalid type (in classification)", item);
						method->d.clos_method.args[argn].type = typeid;
					} else if(item->type == ST_CONS) {
						/* (eql something)? */

						s_obj *iter3 = item;
						s_obj *elem;

						nextparm(&elem, &iter3);
						if(elem->type != ST_SYMBOL)
							return create_error(SE_PARM, "Defmethod: symbol expected as classification type list's first element");
						if(elem->d.symbol.id != _EQL)
							return create_error(SE_PARM, "Defmethod: only EQL is supported as classification type magic");

						nextparm(&elem, &iter3);
						elem = eval(ctx, elem);
						if(elem->type == ST_ERROR) return elem;

						method->d.clos_method.args[argn].type = object_typeid(elem);
						method->d.clos_method.args[argn].value = elem;
					} else {
						/* not a symbol, not a cons. */
						return create_error(SE_PARM, "Defmethod: type in classified pair must be a symbol or a list");
					}
				}

			} else {
				/* argument name not a symbol, not a cons. */
				return create_error(SE_PARM, "Defmethod: argument name must be a symbol or a list");
			}

			argn ++;
		}

		/* set misc args to rest of arglist */
		method->d.clos_method.miscargs = iter;
	} else {
		/* argn == 0 */
		method->d.clos_method.miscargs = args;
	}

	s_obj * res = method_calculate(ctx, name, method);
	if(res->type == ST_ERROR) return res;

	return name;
}

/* register builtins in variable context */
void register_sforms(varctx *ctx)
{
	varctx_set(ctx, "QUOTE", create_sform(sb_quote));

	varctx_set(ctx, "LAMBDA", create_sform_with_data(sb_lambda, (void *) 0));
	varctx_set(ctx, "MACRO", create_sform_with_data(sb_lambda, (void *) 1));
	varctx_set(ctx, "R-MACRO", create_sform_with_data(sb_lambda, (void *) 2));
	varctx_set(ctx, "SYMBOLIC-MACRO", create_sform_with_data(sb_lambda, (void *) 4));
	varctx_set(ctx, "R-SYMBOLIC-MACRO", create_sform_with_data(sb_lambda, (void *) 5));
	varctx_set(ctx, "SUBCTX", create_builtin(sb_subctx));

	varctx_set(ctx, "IF", create_builtin(sb_if));
	varctx_set(ctx, "DO", create_builtin_with_data(sb_do, (void *) 0));
	varctx_set(ctx, "LET", create_builtin_with_data(sb_do, (void *) 1));
	varctx_set(ctx, "CATCH", create_builtin(sb_catch));

	varctx_set(ctx, "DEFMETHOD", create_sform(sb_defmethod));
}

